"""
Daily Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime, timedelta
from typing import Union
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.utilities.aggregations import degree_mean
from meteostat.interface.timeseries import TimeSeries
from meteostat.interface.point import Point


class Daily(TimeSeries):
    """
    Retrieve daily weather observations for one or multiple weather stations or
    a single geographical point
    """

    # The cache subdirectory
    cache_subdir = "daily"

    # Granularity
    granularity = Granularity.DAILY

    # Download data as annual chunks
    # This cannot be changed and is only kept for backward compatibility
    chunked = True

    # Default frequency
    _freq = "1D"

    # Source mappings
    _source_mappings = {
        "dwd_daily": "A",
        "eccc_daily": "A",
        "ghcnd": "B",
        "dwd_hourly": "C",
        "eccc_hourly": "C",
        "isd_lite": "D",
        "synop": "E",
        "dwd_poi": "E",
        "metar": "F",
        "model": "G",
        "dwd_mosmix": "G",
        "metno_forecast": "G",
    }

    # Flag which represents model data
    _model_flag = "G"

    # Columns
    _columns = [
        "year",
        "month",
        "day",
        {"tavg": "temp"},
        "tmin",
        "tmax",
        "prcp",
        {"snow": "snwd"},
        {"wdir": None},
        "wspd",
        "wpgt",
        "pres",
        "tsun",
    ]

    # Index of first meteorological column
    _first_met_col = 3

    # Columns for date parsing
    _parse_dates = ["year", "month", "day"]

    # Default aggregation functions
    aggregations = {
        "tavg": "mean",
        "tmin": "min",
        "tmax": "max",
        "prcp": "sum",
        "snow": "max",
        "wdir": degree_mean,
        "wspd": "mean",
        "wpgt": "max",
        "pres": "mean",
        "tsun": "sum",
    }

    def __init__(
        self,
        loc: Union[pd.DataFrame, Point, list, str],  # Station(s) or geo point
        start=datetime(1781, 1, 1, 0, 0, 0),
        end=datetime.combine(
            datetime.today().date() + timedelta(days=10), datetime.max.time()
        ),
        model=True,  # Include model data?
        flags=False,  # Load source flags?
    ) -> None:
        # Extract relevant years
        if self.chunked:
            self._annual_steps = [
                start.year + i for i in range(end.year - start.year + 1)
            ]
        # Initialize time series
        self._init_time_series(loc, start, end, model, flags)

    def expected_rows(self) -> int:
        """
        Return the number of rows expected for the defined date range
        """

        return (self._end - self._start).days + 1



"""
Core Class - Data Loader

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from io import BytesIO
from gzip import GzipFile
from urllib.request import Request, ProxyHandler, build_opener
from urllib.error import HTTPError
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
from typing import Callable, List, Optional
import pandas as pd
from meteostat.core.warn import warn


def processing_handler(
    datasets: List, load: Callable[[dict], None], cores: int, threads: int
) -> None:
    """
    Load multiple datasets (simultaneously)
    """

    # Data output
    output = []

    # Multi-core processing
    if cores > 1 and len(datasets) > 1:
        # Create process pool
        with Pool(cores) as pool:
            # Process datasets in pool
            output = pool.starmap(load, datasets)

            # Wait for Pool to finish
            pool.close()
            pool.join()

    # Multi-thread processing
    elif threads > 1 and len(datasets) > 1:
        # Create process pool
        with ThreadPool(threads) as pool:
            # Process datasets in pool
            output = pool.starmap(load, datasets)

            # Wait for Pool to finish
            pool.close()
            pool.join()

    # Single-thread processing
    else:
        for dataset in datasets:
            output.append(load(*dataset))

    # Remove empty DataFrames
    filtered = list(filter(lambda df: not df.empty, output))

    return pd.concat(filtered) if len(filtered) > 0 else output[0]


def load_handler(
    endpoint: str,
    path: str,
    proxy: Optional[str] = None,
    names: Optional[List] = None,
    dtype: Optional[dict] = None,
    parse_dates: Optional[List] = None,
    default_df: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """
    Load a single CSV file into a DataFrame
    """

    try:
        handlers = []

        # Set a proxy
        if proxy:
            handlers.append(ProxyHandler({"http": proxy, "https": proxy}))

        # Read CSV file from Meteostat endpoint
        with build_opener(*handlers).open(Request(endpoint + path)) as response:
            # Decompress the content
            with GzipFile(fileobj=BytesIO(response.read()), mode="rb") as file:
                df = pd.read_csv(
                    file,
                    names=names,
                    dtype=dtype,
                    parse_dates=parse_dates,
                )

    except (FileNotFoundError, HTTPError):
        df = default_df if default_df is not None else pd.DataFrame(columns=names)

        # Display warning
        warn(f"Cannot load {path} from {endpoint}")

    # Return DataFrame
    return df



"""
Utilities - DataFrame Mutations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
from typing import Union
import numpy as np
import pandas as pd


def localize(df: pd.DataFrame, timezone: str) -> pd.DataFrame:
    """
    Convert time data to any time zone
    """

    return df.tz_localize("UTC", level="time").tz_convert(timezone, level="time")


def filter_time(
    df: pd.DataFrame,
    start: Union[datetime, None] = None,
    end: Union[datetime, None] = None,
) -> pd.DataFrame:
    """
    Filter time series data based on start and end date
    """

    # Get time index
    time = df.index.get_level_values("time")

    # If no time index, return original DataFrame
    if len(time) == 0:
        return df

    # Filter & return
    return df.loc[(time >= start) & (time <= end)] if start and end else df


def adjust_temp(df: pd.DataFrame, alt: int):
    """
    Adjust temperature-like data based on altitude
    """

    # Default temperature difference by 100 meters
    temp_diff = 0.6

    # Temperature-like columns
    temp_like = ("temp", "dwpt", "tavg", "tmin", "tmax")

    # Adjust values for all temperature-like data
    for col_name in temp_like:
        if col_name in df.columns:
            df.loc[df[col_name] != np.nan, col_name] = df[col_name] + (
                temp_diff * ((df["elevation"] - alt) / 100)
            )

    return df


def calculate_dwpt(df: pd.DataFrame, col: str) -> pd.DataFrame:
    """
    Calculate dew point temperature
    """
    magnus_const_a = 17.27
    magnus_const_b = 237.7  # degrees Celsius

    temperature = df["temp"]
    relative_humidity = df["rhum"]

    alpha = (
        (magnus_const_a * temperature) / (magnus_const_b + temperature)
    ) + np.log(relative_humidity / 100.0)
    df[col] = (magnus_const_b * alpha) / (magnus_const_a - alpha)

    df[col] = df[col].round(1)

    df[f"{col}_flag"] = df[["temp_flag", "rhum_flag"]].max(axis=1, skipna=True)

    return df



"""
Meteorological Data Class

A parent class for both time series and
climate normals data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from collections.abc import Callable
from typing import Dict, List, Union
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.core.loader import processing_handler
from meteostat.utilities.mutations import adjust_temp
from meteostat.utilities.aggregations import weighted_average
from meteostat.interface.base import Base


class MeteoData(Base):
    """
    A parent class for both time series and
    climate normals data
    """

    # The list of weather Stations
    _stations: Union[pd.Index, None] = None

    # The data frame
    _data: pd.DataFrame = pd.DataFrame()

    @property
    def _raw_columns(self) -> List[str]:
        """
        Get the list of raw data columns, excluding any dicts with callable values
        """
        return [
            list(col.values())[0] if isinstance(col, dict) else col
            for col in self._columns
            if not (
                isinstance(col, dict)
                and (
                    isinstance(list(col.values())[0], Callable)
                    or list(col.values())[0] is None
                )
            )
        ]

    @property
    def _processed_columns(self) -> List[str]:
        """
        Get the list of processed data columns, excluding any dicts with callable values
        """
        return [
            list(col.keys())[0] if isinstance(col, dict) else col
            for col in self._columns[self._first_met_col :]
        ]

    @property
    def _renamed_columns(self) -> Dict[str, str]:
        """
        Get the dict of renamed data columns, including `_source` suffixes
        """
        return {
            new_key: new_val
            for d in self._columns
            if isinstance(d, dict)
            for k, v in d.items()
            if not isinstance(v, Callable)
            for new_key, new_val in ((v, k), (f"{v}_source", f"{k}_source"))
        }

    @property
    def _virtual_columns(self) -> Dict[str, str]:
        """
        Get the dict of virtual data columns
        """
        return {
            k: v
            for d in self._columns
            if isinstance(d, dict)
            for k, v in d.items()
            if isinstance(v, Callable)
        }

    def _get_datasets(self) -> list:
        """
        Get list of datasets
        """

        if self.granularity in (Granularity.HOURLY, Granularity.DAILY):
            datasets = [
                (str(station), year)
                for station in self._stations
                for year in self._annual_steps
            ]
        else:
            datasets = [(str(station),) for station in self._stations]

        return datasets

    def _get_data(self) -> None:
        """
        Get all required data dumps
        """

        if len(self._stations) > 0:
            # Get list of datasets
            datasets = self._get_datasets()

            # Data Processings
            return processing_handler(
                datasets, self._load_data, self.processes, self.threads
            )

        # Empty DataFrame
        return pd.DataFrame(columns=self._processed_columns)

    # pylint: disable=too-many-branches
    def _resolve_point(
        self, method: str, stations: pd.DataFrame, alt: int, adapt_temp: bool
    ) -> None:
        """
        Project weather station data onto a single point
        """

        if self._stations.size == 0 or self._data.size == 0:
            return

        if method == "nearest":
            if adapt_temp:
                # Join elevation of involved weather stations
                data = self._data.join(stations["elevation"], on="station")

                # Adapt temperature-like data based on altitude
                data = adjust_temp(data, alt)

                # Drop elevation & round
                data = data.drop("elevation", axis=1).round(1)

            else:
                data = self._data

            if self.granularity == Granularity.NORMALS:
                self._data = data.groupby(level=["start", "end", "month"]).agg("first")

            else:
                self._data = data.groupby(
                    pd.Grouper(level="time", freq=self._freq)
                ).agg("first")

        else:
            # Join score and elevation of involved weather stations
            data = self._data.join(stations[["score", "elevation"]], on="station")

            # Adapt temperature-like data based on altitude
            if adapt_temp:
                data = adjust_temp(data, alt)

            # Exclude non-mean data & perform aggregation
            if not self.granularity == Granularity.NORMALS:
                excluded = data["wdir"]
                excluded = excluded.groupby(
                    pd.Grouper(level="time", freq=self._freq)
                ).agg("first")

            # Aggregate mean data
            if self.granularity == Granularity.NORMALS:
                data = data.groupby(level=["start", "end", "month"]).apply(
                    weighted_average
                )

                # Remove obsolete index column
                try:
                    data = data.reset_index(level=3, drop=True)
                except IndexError:
                    pass

            else:
                data = data.groupby(pd.Grouper(level="time", freq=self._freq)).apply(
                    weighted_average
                )

                # Drop RangeIndex
                data.index = data.index.droplevel(1)

                # Merge excluded fields
                data["wdir"] = excluded

            # Drop score and elevation
            self._data = data.drop(["score", "elevation"], axis=1).round(1)

        # Set placeholder station ID
        self._data["station"] = "XXXXX"

        # Set index
        if self.granularity == Granularity.NORMALS:
            self._data = self._data.set_index("station", append=True)
            self._data = self._data.reorder_levels(["station", "start", "end", "month"])
        else:
            self._data = self._data.set_index(
                ["station", self._data.index.get_level_values("time")]
            )

        # Set station index
        self._stations = pd.Index(["XXXXX"])



"""
Utilities - Helpers

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from typing import Optional
import numpy as np


def get_distance(lat1, lon1, lat2, lon2) -> float:
    """
    Calculate distance between weather station and geo point
    """
    # Earth radius in meters
    radius = 6371000

    # Degress to radian
    lat1, lon1, lat2, lon2 = map(np.deg2rad, [lat1, lon1, lat2, lon2])

    # Deltas
    dlat = lat2 - lat1
    dlon = lon2 - lon1

    # Calculate distance
    arch = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    arch_sin = 2 * np.arcsin(np.sqrt(arch))

    return radius * arch_sin


def _get_flag_from_single_source(
    source: str, source_mappings: dict, model_flag: str
) -> str:
    """
    Get flag from single source
    """
    if source in source_mappings:
        return source_mappings[source]
    return model_flag


def get_flag_from_source_factory(source_mappings: dict, model_flag: str) -> str:
    """
    Get flag from source
    """

    def _get_flag_from_source(source: Optional[str]) -> str:
        sources = source.split(" ")

        flags = [
            _get_flag_from_single_source(src, source_mappings, model_flag)
            for src in sources
        ]

        return "".join(flags)

    return _get_flag_from_source


def with_suffix(items, suffix):
    """
    Takes a list of strings and a suffix, returns a new list containing
    the same items with the suffix added.
    """
    return [item + suffix for item in items]
